import threading
import queue
import asyncio
from collections import defaultdict
import inspect
import random

from uuid import uuid4

from carp.service import (
    ApiFunction,
    ApiMethod,
    ApiNonInstanceMethod,
    CallData,
    CallResponse,
)
from carp.channel import Channel, ChannelError
from carp.serializer import Serializer, Serializable


class RemoteExecutionError(Exception):
    pass


class HostAnnounce(Serializable):
    def __init__(self, *, host_id):
        self.host_id = host_id
        super().__init__()

    def to_dict(self):
        return dict(
            host_id=self.host_id
        )


class HostExports(Serializable):
    def __init__(self, *, host_id, exports, metadata):
        self.host_id = host_id
        self.exports = exports
        self.metadata = metadata
        super().__init__()

    def to_dict(self):
        return dict(
            host_id=self.host_id,
            exports=self.exports,
            metadata=self.metadata,
        )

class HostExitNotify(Serializable):
    def __init__(self, *, host_id):
        self.host_id = host_id

    def to_dict(self):
        return dict(host_id=self.host_id)

class Host:
    STOPPED = "stopped"
    STARTED = "started"

    def __init__(
        self, *,
        label=None,
    ):
        self.id = uuid4().hex
        self.label = label or self.id

        self.services_remote = defaultdict(list)
        self.services_local = {}
        self.services_event = asyncio.Event()
        self.calls_active = {}
        self.hosts_remote = {}
        self.listen_channel = None
        self.task_last_id = 0
        self.tasks = {}
        self.status = Host.STOPPED

        self.event_loop_thread = threading.get_ident()
        self.event_loop = asyncio.get_event_loop()

        # callbacks
        self.event_handlers = defaultdict(list)

    async def _report_services(self):
        host_map = HostExports(
            host_id=self.id,
            exports=list(self.services_local.keys()),
            metadata={
                service_name: service.metadata
                for service_name, service in self.services_local.items()
            }
        ).serialize()
        for channel in self.hosts_remote.values():
            await channel.put(host_map)

    async def _set_status(self, new_status, **kwargs):
        old_status = self.status
        self.status = new_status
        await self.emit("status", old_status, new_status, **kwargs)

    async def _task_wrapper(self, coro, task_id):
        rv = None
        try:
            rv = await coro
        except Exception as e:
            import traceback
            tbinfo = traceback.format_exc()
            await self.emit("exception", e, tbinfo)
        finally:
            if task_id in self.tasks:
                del self.tasks[task_id]
        return rv

    def async_task(self, coro):
        if inspect.isawaitable(coro):
            current_thread = threading.get_ident()
            task_id = self.task_last_id
            self.task_last_id += 1

            if current_thread == self.event_loop_thread:
                task = asyncio.create_task(self._task_wrapper(coro, task_id))
            else:
                task = asyncio.run_coroutine_threadsafe(
                    self._task_wrapper(coro, task_id), self.event_loop
                )
            self.tasks[task_id] = task
            return task
        else:
            return coro

    async def start(self, channel: Channel):
        """
        Start the host, accepting connections on the specified
        Channel
        """
        self.listen_channel = channel
        await channel.serve(on_connect=self.accept)
        await self._set_status(Host.STARTED)

    async def stop(self):
        to_cancel = self.tasks.values()
        self.tasks = {}
        for t in to_cancel:
            t.cancel()
            try:
                await t
            except asyncio.CancelledError:
                pass
        await self._set_status(Host.STOPPED)

    async def accept(self, channel):
        await self.emit("accept", channel)

        await channel.put(HostAnnounce(host_id=self.id).serialize())
        await self._message_loop(channel)
        await channel.close()

    async def connect(self, channel):
        await self._set_status(Host.STARTED)
        await channel.connect()
        await self.emit("connect", channel)

        await channel.put(HostAnnounce(host_id=self.id).serialize())
        self.async_task(self._message_loop(channel))

    async def _message_loop(self, channel):
        async def _process(bdata):
            message = Serializer.deserialize(bdata)

            await self.emit("message", message)

            # process broadcasts by peers
            try:
                if isinstance(message, HostAnnounce):
                    self.hosts_remote[message.host_id] = channel
                    await self._report_services()
                elif isinstance(message, HostExitNotify):
                    for service, hosts in self.services_remote.items():
                        if message.host_id in hosts:
                            hosts.remove(message.host_id)
                elif isinstance(message, HostExports):
                    for service in message.exports:
                        if message.host_id not in self.services_remote[service]:
                            self.services_remote[service].append(message.host_id)
                    await self.emit(
                        "exports", message.host_id,
                        message.exports, message.metadata
                    )
                    self.services_event.set()
                elif isinstance(message, CallData):
                    await self.emit("call", message)
                    service = self.services_local.get(
                        message.service_name.split(".")[0]
                    )
                    if not service:
                        raise RemoteExecutionError(
                            f"Service {message.service_name} not on {self.id}"
                        )
                    response = await self.handle(service, message)
                    await channel.put(response.serialize())
                elif isinstance(message, CallResponse):
                    call_data = self.calls_active.get(message.call_id)
                    # FIXME: Should not be returning CallResponses for
                    # no response methods
                    if call_data:
                        call_data.response = message
                        call_data.event.set()
            except Exception as e:
                for host_id, remote_channel in self.hosts_remote.items():
                    if remote_channel == channel:
                        channel.status = Channel.CLOSED
                        await self.emit("disconnect", host_id)
                        break

        first_message = True
        while (
            self.status == Host.STARTED
            and channel.status == Channel.CONNECTED
        ):
            try:
                message_bytes = await channel.get()
                if first_message:
                    await _process(message_bytes)
                    first_message = False
                else:
                    self.async_task(_process(message_bytes))
            except Exception as e:
                for host_id, remote_channel in self.hosts_remote.items():
                    if remote_channel == channel:
                        channel.status = Channel.CLOSED
                        await self.emit("disconnect", host_id)
                        break
        await channel.close()

    async def export(self, service_impl, metadata=None):
        """
        Announce that a service is available on this host
        """
        api_factory = ApiFunction
        if hasattr(service_impl, '_service_type'):
            api_factory = service_impl._service_type

        service = api_factory(service_impl)
        if metadata:
            service.metadata = metadata

        self.services_local[service.name] = service
        service.is_remote = False
        service.host = self
        service.host_id = self.id
        await self._report_services()
        self.services_event.set()
        return service

    async def require(self, service_impl):
        """
        Use a service announced by this or another host, waiting
        until it is available
        """
        api_factory = ApiFunction
        if hasattr(service_impl, '_service_type'):
            api_factory = service_impl._service_type

        service = api_factory(service_impl)

        while (
            service.name not in self.services_local
            and service.name not in self.services_remote
            and self.status == Host.STARTED
        ):
            await self.services_event.wait()
            self.services_event.clear()

        service.host = self
        if service.name in self.services_local:
            service.is_remote = False
            service.host_id = self.id
        elif service.name in self.services_remote:
            service.host_id = random.choice(self.services_remote[service.name])
            service.is_remote = True

        return service

    async def call(
        self, service, args, kwargs, *,
        response=True, response_queue=None, response_callback=None
    ):
        """
        Send a request to a remote service, waiting for a
        response
        """

        call_data = CallData(
            service_name=service.name,
            host_id=service.host_id,
            instance_id=service.instance_id,
            args=args,
            kwargs=kwargs
        )
        channel = self.hosts_remote[service.host_id]
        if response:
            self.calls_active[call_data.call_id] = call_data

        await channel.put(call_data.serialize())
        if response:
            await call_data.event.wait()
            del self.calls_active[call_data.call_id]

            if call_data.response.exception:
                raise RemoteExecutionError(call_data.response.exception)
            if response_queue:
                response_queue.put(call_data.response.value)
            if response_callback:
                response_callback(call_data.response.value)

            return call_data.response.value

        if response_queue:
            response_queue.put(None)
        return None

    def call_with_cb(self, service, args, kwargs, *, callback=None):
        """
        Non-blocking synchronous call with callback for response
        """
        self.async_task(
            self.call(
                service, args, kwargs, response=True,
                response_callback=callback,
            )
        )
        return None

    def call_blocking(self, service, args, kwargs, response=True):
        """
        Synchronous version of call() for bridging to non-async-aware
        processes
        """
        response_queue = None
        rv = None

        if response:
            response_queue = queue.Queue(1)

        self.async_task(
            self.call(
                service, args, kwargs, response=response,
                response_queue=response_queue,
            )
        )

        if response:
            rv = response_queue.get()
        return rv

    async def handle(self, service, message):
        """
        Handle a remote request using a local service
        """
        call_return = None
        exception = None

        if service.name != message.service_name:
            method_name = message.service_name.split(".", 1)[1]
            class_service = service

            # FIXME should have a better way of getting the right
            # service from the CallData
            if message.instance_id:
                service = ApiMethod(
                    class_service, method_name, message.instance_id
                )
            else:
                service = ApiNonInstanceMethod(class_service, method_name)

        try:
            raw_return = service(*message.args, **message.kwargs)
            if inspect.isawaitable(raw_return):
                call_return = await raw_return
            else:
                call_return = raw_return

        except Exception as e:
            import traceback
            tbinfo = traceback.format_exc()
            await self.emit("exception", e, tbinfo)
            exception = type(e).__name__

        return CallResponse(
            call_id=message.call_id,
            service_name=service.name,
            host_id=message.host_id,
            value=call_return,
            exception=exception,
        )

    def on(self, event_name, handler, uninstall=False):
        prev = self.event_handlers[event_name]
        if not uninstall:
            prev.append(handler)
        else:
            self.event_handlers[event_name] = filter(
                lambda h: h != handler, prev
            )

    async def emit(self, event_name, *args, **kwargs):
        handlers = self.event_handlers[event_name]
        for handler in handlers:
            handled = handler(event_name, *args, **kwargs)
            if inspect.isawaitable(handled):
                handled = await handled
            if handled:
                break

    async def wait_for_completion(self):
        while tasks := self.tasks.values():
            await asyncio.wait(tasks)




from abc import ABC, abstractmethod

class Channel(ABC):
    CLOSED = "closed"
    CONNECTED = "connected"
    CONNECTING = "connecting"
    SERVING = "serving"

    @abstractmethod
    async def serve(self):
        pass

    @abstractmethod
    async def connect(self, **config):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def put(self, message):
        pass

    @abstractmethod
    async def get(self):
        pass






import random

from unittest import TestCase

from carp.serializer import Serializable, Serializer, ProtobufSerializer
from carp.serializer.protobuf.compiled_python.pytypes_pb2 import (
    PythonValueContainer,
    PythonArrayContainer,
    PythonDictContainer
)


class PValue(Serializable):
    serializer = ProtobufSerializer
    protobuf_type = PythonValueContainer

    def __init__(self, value):
        self.value = value


class PArray(Serializable):
    serializer = ProtobufSerializer
    protobuf_type = PythonArrayContainer

    def __init__(self, value):
        self.value = value


class PDict(Serializable):
    serializer = ProtobufSerializer
    protobuf_type = PythonDictContainer

    def __init__(self, value):
        self.value = value


class TestProtobufSerializer(TestCase):
    def test_pyvalue_types(self):
        """
        PythonValue protobuf message works as expected
        """
        values = [
            (123.4,),
            (12,),
            ("hello, world!",),
            (dict(a=1, b=2, c="hello", d=False),),
            ((1,2,3), [1,2,3]),
            (True,),
            (False,),
            (None,),
        ]

        for v in values:
            val = PValue(v[0])
            expected = v[1] if len(v) > 1 else v[0]

            serialized = val.serialize()

            deserialized = Serializer.deserialize(serialized)
            self.assertEqual(expected, deserialized.value)
            self.assertEqual(type(expected), type(deserialized.value))

    def test_python_array(self):
        """
        PythonArray protobuf message works as expected
        """
        val = [
            random.randint(1,1000000) for _ in range(100)
        ]
        pv = PArray(val)

        serialized = pv.serialize()

        deserialized = Serializer.deserialize(serialized)
        self.assertEqual(val, deserialized.value)


    def test_python_dict(self):
        """
        PythonArray protobuf message works as expected
        """
        val = {
            "a": 1,
            "b": 2,
            "c": True,
            "d": None
        }

        pv = PDict(val)

        serialized = pv.serialize()
        deserialized = Serializer.deserialize(serialized)
        self.assertIsInstance(deserialized, PDict)
        self.assertIsInstance(deserialized.value, dict)
        self.assertEqual(val, deserialized.value)



"""
Mixin class to provide hooks for Carp to serialize and deserialize
objects

Default implementation uses JSON
"""


class Serializable:
    registry = {}
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Serializable.registry[cls.__name__] = cls

    def serialize(self):
        from .json_serializer import JsonSerializer
        # defaults to JsonSerializer
        serializer_type = getattr(self, 'serializer', None) or JsonSerializer
        return serializer_type().serialize(self)
    





class Serializer:
    serializers = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Serializer.serializers[cls.label] = cls

    @staticmethod
    def serialize(obj):
        from .serializable import Serializable
        from .json_serializer import JsonSerializer
        if isinstance(obj, Serializable):
            return obj.serialize()

        return JsonSerializer().serialize(obj)

    @staticmethod
    def deserialize(bytestr: bytes) -> any:
        serializer_label, data = bytestr.split(b":", 1)
        serializer = Serializer.serializers.get(serializer_label.decode())
        if serializer:
            return serializer().deserialize(data)
        raise ValueError(f"Serializer '{serializer_label}' not found: {bytes.decode('utf-8')}")



import tempfile
from carp.channel import UnixSocketChannel
from carp.service import apiclass, ApiClass, ApiProxyObject, ApiNonInstanceMethod
from carp.host import Host, RemoteExecutionError

from unittest import IsolatedAsyncioTestCase


@apiclass
class RemoteObj:
    def __init__(self):
        self.counter = 0

    def set_counter(self, val):
        self.counter = val

    def get_counter(self):
        return self.counter

    @staticmethod
    def static_value():
        return "Hello, world!"

    @classmethod
    def class_value(cls):
        return cls.__name__

class TestApiClass(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.sockname = tempfile.mktemp()
        self.server_channel = UnixSocketChannel(socket_path=self.sockname)
        self.server_host = Host()
        self.client_channel = UnixSocketChannel(socket_path=self.sockname)
        self.client_host = Host()

        await self.server_host.start(self.server_channel)
        await self.client_host.connect(self.client_channel)

    async def asyncTearDown(self):
        await self.client_host.stop()
        await self.server_host.stop()
        await self.client_channel.close()
        await self.server_channel.close()

    async def test_create_obj(self):
        await self.server_host.export(RemoteObj)

        RemoteObjFactory = await self.client_host.require(RemoteObj)
        self.assertEqual(RemoteObjFactory.is_remote, True)

        oo = await RemoteObjFactory()
        self.assertEqual(oo._service.is_remote, True)

        self.assertEqual(type(RemoteObjFactory), ApiClass)
        self.assertEqual(type(oo), ApiProxyObject)

    async def test_change_state(self):
        await self.server_host.export(RemoteObj)

        RemoteObjFactory = await self.client_host.require(RemoteObj)
        oo = await RemoteObjFactory()

        self.assertEqual(await oo.get_counter(), 0)
        await oo.set_counter(99)
        self.assertEqual(await oo.get_counter(), 99)

    async def test_static_method(self):
        await self.server_host.export(RemoteObj)

        RemoteObjFactory = await self.client_host.require(RemoteObj)
        vv = await RemoteObjFactory.static_value()
        self.assertEqual(vv, "Hello, world!")

    async def test_class_method(self):
        await self.server_host.export(RemoteObj)

        RemoteObjFactory = await self.client_host.require(RemoteObj)
        self.assertEqual(RemoteObjFactory.is_remote, True)

        cmeth = RemoteObjFactory.class_value
        self.assertEqual(type(cmeth), ApiNonInstanceMethod)
        self.assertEqual(cmeth.is_remote, True)

        vv = await RemoteObjFactory.class_value()
        self.assertEqual(vv, "RemoteObj")




"""
service.py -- Service base class
"""


class Service:
    def __init__(self, name):
        self.name = name
        self.host = None
        self.host_id = None
        self.instance_id = None
        self.is_remote = False
        self.metadata = {}

    @staticmethod
    def build(impl_object):
        factory = None
        if hasattr(impl_object, '_service_type'):
            factory = impl_object._service_type
        
        return factory(impl_object)



import asyncio
import tempfile
from carp.channel import UnixSocketChannel
from carp.service import ApiFunction
from carp.host import Host, RemoteExecutionError

from unittest import IsolatedAsyncioTestCase


class TestHostCall(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.sockname = tempfile.mktemp()
        self.server_channel = UnixSocketChannel(socket_path=self.sockname)
        self.server_host = Host()
        self.client_channel = UnixSocketChannel(socket_path=self.sockname)
        self.client_host = Host()

        await self.server_host.start(self.server_channel)
        await self.client_host.connect(self.client_channel)

    async def asyncTearDown(self):
        await self.client_host.stop()
        await self.server_host.stop()
        await self.server_channel.close()
        await self.client_channel.close()

    async def test_call_locally(self):
        """
        A service that is local to the caller can be used
        """
        func_called = asyncio.Event()
        async def svc_func():
            func_called.set()
            return True

        await self.client_host.export(svc_func)
        clnt_svc = await self.client_host.require(svc_func)

        result = await clnt_svc()

        self.assertEqual(result, True)

    async def test_call_serv_to_client_remotely(self):
        """
        A service that is remote from the caller can be used
        """
        func_called = asyncio.Event()

        async def svc_func():
            func_called.set()
            return True

        await self.client_host.export(svc_func)
        server_svc = await self.server_host.require(svc_func)

        result = await server_svc()
        self.assertEqual(result, True)

    async def test_call_client_to_serv_remotely(self):
        """
        A service that is remote from the caller can be used
        """
        func_called = asyncio.Event()

        async def svc_func():
            func_called.set()
            return True

        await self.server_host.export(svc_func)
        client_svc = await self.client_host.require(svc_func)

        result = await client_svc()
        self.assertEqual(result, True)

    async def test_call_remotely__args(self):
        """
        A service with arguments works as expected
        """
        func_called = asyncio.Event()

        async def svc_func(arg1, arg2):
            func_called.set()
            return [arg1, arg2]

        await self.server_host.export(svc_func)
        client_svc = await self.client_host.require(svc_func)

        test_values = [
            [12, 34],
            ["hello", "world"],
            [[12, 34, 56], [78, 90]],
            [dict(a=1, b=2), dict(c=3, d=4)]
        ]

        for v in test_values:
            result = await client_svc(*v)
            self.assertEqual(result, v)

    async def test_call_remotely__exception(self):
        """
        Remote service raising an exception raises one locally
        """
        func_called = asyncio.Event()

        async def svc_func():
            func_called.set()
            raise ValueError

        await self.server_host.export(svc_func)
        client_svc = await self.client_host.require(svc_func)

        with self.assertRaises(RemoteExecutionError):
            await client_svc()




from .service import Service

def apifunc(func: callable) -> callable:
    """
    Decorator for standalone API functions
    """
    func._service_type = ApiFunction
    func._service_name = func.__name__
    return func


class ApiFunction(Service):
    def __init__(self, func: callable):
        self.func = func
        self.host = None
        name = func.__name__
        if hasattr(func, '_service_name'):
            name = func._service_name
        super().__init__(name)

    async def __call__(self, *args, **kwargs):
        needs_response = getattr(self.func, "needs_response", True)

        if self.is_remote:
            rv = await self.host.call(self, args, kwargs, response=needs_response)
        else:
            rv = await self.func(*args, **kwargs)
        return rv



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: call_data.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import pytypes_pb2 as pytypes__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x63\x61ll_data.proto\x12\x04\x63\x61rp\x1a\rpytypes.proto\"\x9a\x01\n\x08\x43\x61llData\x12\x0f\n\x07\x63\x61ll_id\x18\x01 \x01(\x03\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x13\n\x0binstance_id\x18\x03 \x01(\x03\x12\x14\n\x0cservice_name\x18\x04 \x01(\t\x12\x1f\n\x04\x61rgs\x18\x05 \x01(\x0b\x32\x11.carp.PythonArray\x12 \n\x06kwargs\x18\x06 \x01(\x0b\x32\x10.carp.PythonDict\"{\n\x0c\x43\x61llResponse\x12\x0f\n\x07\x63\x61ll_id\x18\x01 \x01(\x03\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x14\n\x0cservice_name\x18\x03 \x01(\t\x12 \n\x05value\x18\x04 \x01(\x0b\x32\x11.carp.PythonValue\x12\x11\n\texception\x18\x05 \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'call_data_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CALLDATA._serialized_start=41
  _CALLDATA._serialized_end=195
  _CALLRESPONSE._serialized_start=197
  _CALLRESPONSE._serialized_end=320
# @@protoc_insertion_point(module_scope)
